/*****
*
* This file is part of the OMS program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

/****************************************************************************
 * recycler.c -- Gestion d'un espace de stockage recyclable.	            *
 *                                                                          *
 * (c) 2000 Bertrand Petits, tous droits reserves.                          *
 ****************************************************************************/

#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <limits.h>
#include <assert.h>

#include "oms/recycler.h"

typedef struct {
	struct desc *descriptor;
} chunk_t;

/*
 * FIXME: do this using a structure (!typedef)
 * and an extenssible field.
 */
typedef struct {
	struct desc *descriptor;
	union {
		void *aligned1;
		char aligned2;
		short aligned3;
		int aligned4;
		long aligned5;
		void (*aligned6) (void);
	} aligned;
} alignedchunk_t;

/*
 * Element descriptor
 */
typedef struct desc {
	chunk_t *chunk;
	recycler_t *recycler;
	unsigned int refcount;
	struct desc *next, *prev, *self;
} desc_t;

/*
 * Definition of the recycler_t type;
 * The handle given to the user is a pointer on this structure.
 */
struct _recycler {
	size_t chunksize;
	unsigned int count, max;
	desc_t *freelist, *usedlist;
};

/*
 * Local variable.
 */
static const int UserDataOffset = offsetof(alignedchunk_t, aligned);

/*
 * Create a new element for a recycler.
 *
 * We return a pointer on the descriptor of this
 * new element or NULL in case we fail.
 *
 * The prev and next field value of the new descriptor are unpredictable.
 */
static desc_t *CreateNewChunk(recycler_t *Recycler)
{
	desc_t *Descriptor;

        /*
         * Allocate the element descriptor.
         */
	Descriptor = (desc_t *) malloc(sizeof(desc_t));
	if (Descriptor == NULL)
		return NULL;
	Descriptor->self = Descriptor;

        /*
         * Allocate the element memory, preceded by a structure * chunk_t.
         * We take care that the user data are aligned as the compiler recommand.
         */
	Descriptor->chunk =
	    (chunk_t *) malloc(Recycler->chunksize + UserDataOffset);
	if (Descriptor->chunk == NULL) {
		free(Descriptor);
		return (NULL);
	}
	Descriptor->chunk->descriptor = Descriptor;

        /*
         * The descriptor point on the recycler.
         * This is not used at the moment, but should be in a few time.
         */
	Descriptor->recycler = Recycler;

        /*
         * This new element is not referenced yet.
         */
	Descriptor->refcount = 0;

        /*
         * We return a pointer on the descriptor.
         */
	return Descriptor;
}

/*
 * Create a recycler.
 *
 * It will be capable to handle ChunksMax elements of size ChunkSize bytes.
 * Return an handle or NULL for error (see errno in this case.
 */
recycler_t *RecyclerCreate(size_t ChunkSize, unsigned int ChunksMax)
{
	recycler_t *Recycler;
	unsigned int i;
	desc_t *Descriptor, *LastDescriptor;

        /*
         * Create the structure for the recycler.
         */
	Recycler = (recycler_t *) malloc(sizeof(recycler_t));
	if (Recycler == NULL)
		return (NULL);

	/*
         * Initialize it.
         */
	Recycler->chunksize = ChunkSize;
	Recycler->count = 0;
	Recycler->max = ChunksMax;
	Recycler->usedlist = NULL;

        /*
         * Allocate one or all elements & their descriptors and link
         * them together. We could have done only one malloc() for the
         * elements memory and another for the descriptors, byt
         * this is not acceptable because we can't handle this necessary
         * alignment for this system.
         * We leave it to malloc().
         */

        /*
         * Create the first element & keep it's descriptor in the recycler.
	 */
	Descriptor = CreateNewChunk(Recycler);
	if (Descriptor == NULL) {
		free(Recycler);
		return (NULL);
	}
	Recycler->freelist = Descriptor;
	Descriptor->prev = NULL;
	LastDescriptor = Descriptor;

        /*
         * Create the others elements, and link their descriptors in a linked list.
         */
	for (i = 1; i < ChunksMax; i++) {
              
                
                Descriptor = CreateNewChunk(Recycler);
		if (Descriptor == NULL) {
			RecyclerDestroy(Recycler, FALSE);
			return (NULL);
		}
		Descriptor->prev = LastDescriptor;
		LastDescriptor->next = Descriptor;
		LastDescriptor = Descriptor;
	}
	Descriptor->next = NULL;

	return (Recycler);
}

/*
 * Liberate all elements and their descriptors associated in a list desc_t.
 */
static void FreeChunksList(desc_t * List)
{
	desc_t *Next;

	while (List != NULL) {
		Next = List->next;

		free(List->chunk);
		free(List);
		List = Next;
	}
}

/*
 * Liberate the memory used by a recycler.
 * After this function return, the passed handle is invalid.
 *
 * Return 0 for success,
 * -1 for error, with errno == EINVAL(bad handle),
 * ENOTEMPTY(some chunk were not released).
 */
int RecyclerDestroy(recycler_t * Recycler, int Force)
{
      	if (Recycler == NULL) {
		errno = EINVAL;
		return -1;
	}

        /*
         * If some elements in the recycler weren't freed, and that the
         * liberation of theses elements isn't forced, we return an error.
         * else, we free theses elements.
         */
	if (Recycler->usedlist != NULL) {
		if (!Force) {
			errno = ENOTEMPTY;
			return (-1);
		}
		FreeChunksList(Recycler->usedlist);
	}

        /*
         * Free the 'free' elements.
         */
	FreeChunksList(Recycler->freelist);
        
	free(Recycler);

	return 0;
}

/*
 * Add ChunksCount new free elements to a recycler.
 * Return zero on succes, or -1 on error (see errno).
 */
int RecyclerGrow(recycler_t * Recycler, unsigned int ChunksCount)
{
	desc_t *NewList, *NewDescriptor, *BottomDescriptor;
	unsigned int i;

	if (Recycler == NULL) {
		errno = EINVAL;
		return -1;
	}

        /*
         * Verify that the number of elements to add is valid:
         * We need at least one, and we can't overflow the counter limit.
         */
	if (ChunksCount < 1 ||
	    ChunksCount + Recycler->max <= ChunksCount ||
	    ChunksCount + Recycler->max <= Recycler->max) {
		errno = EINVAL;
		return -1;
	}

        /*
         * The new elements are created one by one,
         * We put them together in an independent linked list.
         * This list will be linked to the free elements list of the descriptor
         * only if we can allocate all the required elements.
         */

        /*
         * Create the first of the new elements.
         */
        
	BottomDescriptor = NewList = CreateNewChunk(Recycler);
	if (NewList == NULL)
		return -1;
	NewList->next = NULL;

	for (i = 1; i < ChunksCount; i++) {
		NewDescriptor = CreateNewChunk(Recycler);
		if (NewDescriptor == NULL) {
			FreeChunksList(NewList);
			return -1;
		}
                
		NewDescriptor->next = NewList;
		NewList = NewList->prev = NewDescriptor;
	}

        /*
         * Allocations ended, we join the new list with the free elements list
         * of the the recycler.
         */
	BottomDescriptor->next = Recycler->freelist;
	Recycler->freelist->prev = BottomDescriptor;
	Recycler->freelist = NewList;
        
	Recycler->max += ChunksCount;

	/*
         * Grow succeed.
         */
	return 0;
}

/*
 * Add ChunksCount new free elements to a recycler.
 * Return 0 for success, -1 for error (see errno).
 *
 * Be carefull: the element returned by the last RecyclerGetChunk() call
 * will be free if it wasn't locked.
 */
int RecyclerShrink(recycler_t *Recycler, unsigned int ChunksCount)
{
	unsigned int FreeCount = Recycler->max - Recycler->count, i;
	desc_t *NewList;
        
	if (Recycler == NULL) {
		errno = EINVAL;
		return -1;
	}

        /*
         * Is the number of element we want to add is valid ?
         */
	if (ChunksCount > FreeCount) {
		errno = EINVAL;
		return -1;
	}

        /*
         * Particular case :
         * We need to free all non locked elements.
         */
	if (ChunksCount == FreeCount) {
		FreeChunksList(Recycler->freelist);
		Recycler->freelist = NULL;
		Recycler->max -= ChunksCount;

                return 0;
	}


        /*
         * We extract the first (ChunksCount) members of the lists
         * of non locked elements in an independant list which is freed.
	 */
	NewList = Recycler->freelist;
	for (i = 0; i < ChunksCount; i++)
		NewList = NewList->next;
	NewList->prev->next = NULL;
	FreeChunksList(Recycler->freelist);

        /*
         * The list of non allocated elements is no truncated.
         */
	Recycler->freelist = NewList;
	Recycler->max -= ChunksCount;

	return 0;
}

/*
 * Return a pointer on an element directly usable by the user or NULL,
 * if there's no more elements.
 * See errno in case of error.
 *
 * The returned element isn't locked:
 * a second call to RecyclerGetChunk() without a call to RecyclerLockChunk()
 * will return the same element.
 */
void *RecyclerGetChunk(recycler_t * Recycler)
{
	if (Recycler == NULL) {
		errno = EINVAL;
		return NULL;
	}

        /*
         * If no element is availlable generate an error.
         */
	if (Recycler->freelist == NULL) {
		errno = ENOMEM;
		return NULL;
	}

        /*
         * Return the pointer on the chunk as it must be seen by the user.
         */
	return ((char *) Recycler->freelist->chunk + UserDataOffset);
}

int RecyclerIsLocked(const void *Data) 
{
        desc_t *Descriptor;
	recycler_t *Recycler;
        
	if (Data == NULL) {
		errno = EINVAL;
		return -1;
	}

        /*
         * Get the descriptor using the pointer we got.
         */
	Descriptor = ((const chunk_t *)
                      ((const char *) Data - UserDataOffset))->descriptor->self;
	Recycler = Descriptor->recycler;

        if ( Descriptor->refcount > 0 )
                return TRUE;

        return FALSE;
}


/*
 * Lock a recycler element to avoid it to be reused.
 */
int RecyclerLockChunk(const void *Data)
{
	desc_t *Descriptor;
	recycler_t *Recycler;
                
	if (Data == NULL) {
		errno = EINVAL;
		return -1;
	}

        /*
         * Get the descriptor using the pointer we got.
         */
	Descriptor = ((const chunk_t *)
                      ((const char *) Data - UserDataOffset))->descriptor->self;
	Recycler = Descriptor->recycler;

        /*
         * If the element isn't locked yet, move it's descriptor
         * to the list of used descriptor.
         */
	if (Descriptor->refcount == 0) {
                /*
                 * Remove the descriptor from the free descriptor list.
                 */
		{
			desc_t *Next = Descriptor->next,
			    *Prev = Descriptor->prev;

			if (Next == NULL && Prev == NULL)
				Recycler->freelist = NULL;
			else {
				if (Next != NULL)
					Next->prev = Prev;
				if (Prev != NULL)
					Prev->next = Next;
				else
					Recycler->freelist = Next;
			}
		}

                /*
                 * Add the descriprot to the list head of the used descriptor list.
                 */
		if (Recycler->usedlist == NULL) {
			Recycler->usedlist = Descriptor;
			Descriptor->prev = Descriptor->next = NULL;
		} else {
			Recycler->usedlist->prev = Descriptor;
			Descriptor->next = Recycler->usedlist;
			Descriptor->prev = NULL;
			Recycler->usedlist = Descriptor;
		}

                /*
                 * this is a new locked element.
                 */
		Recycler->count++;
	}

        /*
         * This element is referenced in one more place.
         */
	Descriptor->refcount++;

	return 0;
}

/*
 * Liberate an element from recycler for it to be reused.
 */
int RecyclerReleaseChunk(const void *Data)
{
	desc_t *Descriptor;
	recycler_t *Recycler;
        
	if (Data == NULL) {
		errno = EINVAL;
		return -1;
	}

        /*
         * get the descriptor and the recycler using the pointer
         * we got.
         */
	Descriptor =
	    ((const
	      chunk_t *) ((const char *) Data -
			  UserDataOffset))->descriptor->self;
	Recycler = Descriptor->recycler;

        /*
         * If the element is already locked after we decrement refcount,
         * return
         */
	if (--Descriptor->refcount != 0)
		return 0;

        /*
         * The descriptor is now unused,
         * we remove it from the used descriptor lists.
         */
	{
		desc_t *Next = Descriptor->next, *Prev = Descriptor->prev;

		if (Next == NULL && Prev == NULL)
			Recycler->usedlist = NULL;
		else {
			if (Next != NULL)
				Next->prev = Prev;
			if (Descriptor->prev != NULL)
				Prev->next = Next;
			else
				Recycler->usedlist = Next;
		}
	}

        /*
         * Add the descriptor to the begining of the free descriptors list.
         */
        
	if (Recycler->freelist == NULL) {
		Recycler->freelist = Descriptor;
		Descriptor->prev = Descriptor->next = NULL;
	} else {
		Recycler->freelist->prev = Descriptor;
		Descriptor->next = Recycler->freelist;
		Descriptor->prev = NULL;
		Recycler->freelist = Descriptor;
	}

        /*
         * A locked element less.
         */
	Recycler->count--;
        
	return 0;
}


unsigned int RecyclerGetChunksMax(recycler_t * Recycler)
{
	if (Recycler == NULL) {
		errno = EINVAL;
		return 0;
	}

	return Recycler->max;
}

unsigned int RecyclerGetChunksCount(recycler_t * Recycler)
{
        if (Recycler == NULL) {
		errno = EINVAL;
		return UINT_MAX;
	}

	return Recycler->count;
}

int RecyclerHasFreeChunks(recycler_t * Recycler)
{
        if (Recycler == NULL) {
		errno = EINVAL;
		return -1;
	}

	return (Recycler->freelist != NULL);
}

