
/*
 *
 * Copyright (C) 1999-2000  Thomas Mirlacher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <byteswap.h>

#include <dvb/avp.h>
#include <dvb/pes.h>
#include <oms/oms.h>
#include <oms/buf.h>
#include <oms/plugin.h>


static int _avp_init (plugin_codec_t *plugin, buf_t *buf, char *output_path);
static int _avp_open (plugin_codec_t *plugin);
static int _avp_read (plugin_codec_t *plugin, buf_t *buf, buf_entry_t *buf_entry);
static int _avp_set_stat (plugin_codec_t *plugin, u_int status);
static int _avp_set_attr (plugin_codec_t *plugin, u_int id, void *attr);


static struct avp_priv_struct {
	pthread_t loop;
	buf_t *buf;

	int fd;

	struct {
		u_char cont_count;
		u_char stuff_len;
		u_char stuff_bytes[4];
	} avp[_BUF_MAX];
} avp_priv = {
	0,
	NULL,
	-1,
	{ 0, 0, {0, 0, 0, 0,}},
};

static int _avp_encaps (struct avp_priv_struct *priv, buf_entry_t *buf_entry);

static plugin_codec_t codec_sw = {
	&avp_priv,
	_avp_init,
	NULL,
	_avp_open,
	NULL,
	_avp_read,	// read
	NULL,		// write
	_avp_set_stat,
	_avp_set_attr,
}; 


/**
 *
 **/

static int _avp_read (plugin_codec_t *plugin, buf_t *buf, buf_entry_t *buf_entry)
{
	struct avp_priv_struct *priv = plugin->priv;
	u_int size;

	switch (buf_entry->buf_id) {
		case BUF_VIDEO:
			_avp_encaps (priv, buf_entry);
			break;
		case BUF_AUDIO:
		default:
			break;
	}	

	size = buf_entry->data_len;
	while (size)
		size -= write (priv->fd, buf_entry->data, size);

	return 0;
}


/**
 *
 **/

static int _avp_init (plugin_codec_t *plugin, buf_t *buf, char *output_path)
{
	struct avp_priv_struct *priv = plugin->priv;

	priv->buf = buf;
	
	fprintf (stderr, "avp_init\n");
	priv->fd = open (output_path, O_WRONLY | O_CREAT);

	return 0;
}


/**
 *
 **/

static int _avp_set_stat (plugin_codec_t *plugin, u_int status)
{
        switch (status) {
        	case STATUS_PLAY:
        	case STATUS_STOP:
                	break;
        }

        return 0;
}


/**
 *
 **/

static int _avp_set_attr (plugin_codec_t *plugin, u_int id, void *attr)
{
        switch (id) {
        	case BUF_AUDIO:
        	case BUF_VIDEO:
        	case BUF_SUBPIC:
                	break;
        }

        return 0;
}


/**
 *
 **/

static int _avp_encaps (struct avp_priv_struct *priv, buf_entry_t *buf_entry)
{
	u_int hdr_len = AVP_HDR_LEN;
	u_int len = buf_entry->data_len;
	u_int stuff_len = 0;
	avp_hdr_t *avp;

	u_int offset = priv->avp[buf_entry->buf_id].stuff_len;

	if (buf_entry->flags & BUF_FLAG_PTS_VALID)
		offset += 4;	// PTS len

	hdr_len += offset;

	len += offset;
	stuff_len = len % 4;
	len -= stuff_len;

	if ((buf_entry->data - buf_entry->mem) < hdr_len) {
		fprintf (stderr, "_avp_encaps: need more headroom\n");
		return -1;
	}

// move saved bytes (not 4 byt aligned from previous pkt) into this stream
	memcpy (buf_entry->data-offset, priv->avp[buf_entry->buf_id].stuff_bytes, priv->avp[buf_entry->buf_id].stuff_len);

	avp = (avp_hdr_t *) (buf_entry->data - hdr_len);

	avp->prefix1 = 'A'; 
	avp->prefix2 = 'V';
	avp->stream_id = (buf_entry->buf_id == BUF_VIDEO) ? 0x01 : 0x02;
	avp->cont_count = priv->avp[buf_entry->buf_id].cont_count++;
	avp->reserved = 0x55;

	avp->nothing = 0;

	if (buf_entry->flags & BUF_FLAG_PTS_VALID) {
		u_int32_t *pts = (u_int32_t *) (avp) + AVP_HDR_LEN4;

		avp->stuff_pre = priv->avp[buf_entry->buf_id].stuff_len;
		avp->pts_present = 1;
		*pts = bswap_32 (buf_entry->pts);
	} else {
		avp->stuff_pre = 0;
		avp->pts_present = 0;
	}
	avp->stuff_post = 0;

	avp->length_hi = len >> 8;
	avp->length_lo = len;

// save data which is not 4 byte aligned
	memcpy (priv->avp[buf_entry->buf_id].stuff_bytes, buf_entry->data+buf_entry->data_len-stuff_len, stuff_len);
	priv->avp[buf_entry->buf_id].stuff_len = stuff_len;

	buf_entry->data_len = len + AVP_HDR_LEN;
	buf_entry->data -= hdr_len;

	return 0;
}


/**
 *
 **/

void *plugin_init (char *whoami)
{
	pluginRegister (whoami,
		PLUGIN_ID_CODEC_VIDEO,
		STREAM_ID_MPEG2,
		&codec_sw);

	pluginRegister (whoami,
		PLUGIN_ID_CODEC_AUDIO,
		STREAM_ID_MPEG2,
		&codec_sw);

	return &codec_sw;
}


/**
 *
 **/

static int _avp_open (plugin_codec_t *plugin)
{
	// open fd here ...
}


